#pragma once

#include<dune/phasefield/localoperator/fem_base_chns.hh>

template<typename RF, const int dim, typename CHLFS>
class CHVariables_Threephase_Micro_Extension : public CHVariables_Threephase_Extension<RF,dim,CHLFS>
{
public:

  const typename CHLFS::template Child<4>::Type& vyspace;

  RF vy;
  Dune::FieldVector<RF,dim> gradvy;


  template<typename Pos, typename Cache, typename LV, typename S>
  CHVariables_Threephase_Micro_Extension(const Cache& cache,
                            const Pos& position, const CHLFS& chLfs, const LV& lv, S& s) :
    CHVariables_Threephase_Extension<RF,dim,CHLFS>(cache, position, chLfs, lv, s),
    vyspace(child(chLfs,Dune::Indices::_4)),
    vy(0.0),
    gradvy(0.0)
  {
    for (size_t i=0; i<this->pfspace.size(); i++)
    {
      vy += lv[vyspace.localIndex(i)]*this->basis.phi[i];
      gradvy.axpy(lv[vyspace.localIndex(i)],this->basis.gradphi[i][0]);
    }
  }
};


//_________________________________________________________________________________________________________________________

template<typename RF, const int dim, typename NSLFS>
class NSMicroVariables
{
public:
  const typename NSLFS::template Child<0>::Type& vxspace;
  BasisFN<typename NSLFS::template Child<0>::Type> vbasis;

  RF vx;
  Dune::FieldVector<RF,dim> gradvx;

  template<typename Pos, typename vCache, typename LV, typename S>
  NSMicroVariables(const vCache& vcache, const Pos& position, const NSLFS& nsLfs, const LV& lv, S& s) :
    vxspace(child(nsLfs,Dune::Indices::_0)),
    vbasis(vcache, position, vxspace, s),
    vx(0.0),
    gradvx(0.0)
  {
    for (size_t i=0; i<this->vxspace.size(); i++)
    {
      vx += lv[vxspace.localIndex(i)]*vbasis.phi[i];
      gradvx.axpy(lv[vxspace.localIndex(i)],vbasis.gradphi[i][0]);
    }
  }
};


template<typename NSGFS>
using Micro_VXLocalBasisCache =
  Dune::PDELab::LocalBasisCache<typename NSGFS::template Child<0>::Type::Traits::FiniteElementType::Traits::LocalBasisType>;
