#pragma once

#include<dune/pdelab/common/quadraturerules.hh>
#include<dune/pdelab/finiteelement/localbasiscache.hh>
#include<dune/pdelab/localoperator/defaultimp.hh>
#include<dune/pdelab/localoperator/pattern.hh>
#include<dune/pdelab/localoperator/flags.hh>
#include<dune/pdelab/localoperator/variablefactories.hh>
#include<dune/pdelab/gridfunctionspace/gridfunctionspace.hh>

#include"fem_base_micro.hh"

//Parameter class (for functions), FEM space and RangeField
template<typename Param, typename CHGFS, typename CHContainer, typename NSGFS, typename NSContainer>
class TMIXED_CHNS_R_CahnHilliard :
  public Dune::PDELab::FullVolumePattern,
  public Dune::PDELab::LocalOperatorDefaultFlags,
  public Dune::PDELab::NumericalJacobianVolume<TMIXED_CHNS_R_CahnHilliard<Param, CHGFS, CHContainer, NSGFS, NSContainer> >,
  public Dune::PDELab::NumericalJacobianApplyVolume<TMIXED_CHNS_R_CahnHilliard<Param, CHGFS, CHContainer, NSGFS, NSContainer> >
{
private:
  typedef typename CHContainer::field_type RF;
  typedef typename CHGFS::Traits::GridViewType::template Codim<0>::Entity   ENTITY;

  CHLocalBasisCache<CHGFS> chCache;
  Micro_VXLocalBasisCache<NSGFS> vCache;

  typedef GFS_LocalViewer<CHGFS, CHContainer, ENTITY, RF> CHLView;
  mutable CHLView chOldLocal;
  mutable CHLView chUpwindLocal;

  typedef GFS_LocalViewer<NSGFS, NSContainer, ENTITY, RF> NSLView;
  mutable NSLView nsLocal;
  mutable NSLView nsUpwindLocal;

  Param& param; // parameter functions
  RF& dxp;
  RF& dxpUpwind;
  RF& c;
  const RF dx;
  const RF xpos;

public:
  enum {doPatternVolume=true};
  //enum {doLambdaVolume=true};
  enum {doAlphaVolume=true};

  //constructor
  TMIXED_CHNS_R_CahnHilliard(Param& param_, CHGFS& chGfs_, CHContainer& chVOld_, CHContainer& chVUpwind_,
      NSGFS& nsGfs_, NSContainer& nsV, NSContainer& nsVUpwind_, RF& dxp_, RF& dxpUpwind_, RF& c_, RF& dx_) :
    chOldLocal(chGfs_,chVOld_),
    chUpwindLocal(chGfs_,chVUpwind_),
    nsLocal(nsGfs_,nsV),
    nsUpwindLocal(nsGfs_,nsVUpwind_),
    param(param_),
    dxp(dxp_),
    dxpUpwind(dxpUpwind_),
    c(c_),
    dx(dx_),
    xpos(param.TempXpos)
  {
  }


  template<typename EG, typename LFSU, typename X, typename LFSV, typename R>
  void alpha_volume (const EG& eg, const LFSU& lfsu, const X& x, const LFSV& lfsv, R& r) const
  {
    //Quadrature Rule
    const int dim = EG::Entity::dimension;
    const int order = 2; //TODO: Choose more educated
    auto rule = Dune::PDELab::quadratureRule(eg.geometry(), order);

    chOldLocal.DoEvaluation(eg);
    chUpwindLocal.DoEvaluation(eg);
    nsLocal.DoEvaluation(eg);
    nsUpwindLocal.DoEvaluation(eg);

    for(const auto& ip : rule)
    {
      const auto S = eg.geometry().jacobianInverseTransposed(ip.position());
      auto factor = ip.weight()*eg.geometry().integrationElement(ip.position());

      CHVariables_Threephase_Micro_Extension<RF,dim,LFSV>
        ch(chCache, ip.position(), lfsv, x.base(), S);
      CHVariables_Threephase_Micro_Extension<RF,dim,decltype(chOldLocal.lfs)>
        chOld(chCache, ip.position(),chOldLocal.lfs, chOldLocal.lv, S);
      CHVariables_Threephase_Micro_Extension<RF,dim,decltype(chUpwindLocal.lfs)>
        chUpwind(chCache, ip.position(),chUpwindLocal.lfs, chUpwindLocal.lv, S);

      NSMicroVariables<RF,dim,decltype(nsLocal.lfs)>
        ns(vCache, ip.position(), nsLocal.lfs, nsLocal.lv, S);
      NSMicroVariables<RF,dim,decltype(nsUpwindLocal.lfs)>
        nsUpwind(vCache, ip.position(), nsUpwindLocal.lfs, nsUpwindLocal.lv, S);

//TODO



      RF pf3 = 1-ch.pf-ch.pf2;
      RF pf_f=ch.pf+ch.pf2+2*param.delta*pf3;
      RF pf3Upwind = 1-chUpwind.pf-chUpwind.pf2;
      RF pf_fUpwind = chUpwind.pf+chUpwind.pf2+2*param.delta*pf3Upwind;
      RF pf3Old = 1-chOld.pf-chOld.pf2;
      RF pf_fOld=chOld.pf+chOld.pf2+2*param.delta*pf3Old;

      auto gradpf3 = -ch.gradpf - ch.gradpf2;
      auto gradpf_f = ch.gradpf + ch.gradpf2+2*param.delta*gradpf3;

      auto gradpf3Old = -chOld.gradpf - chOld.gradpf2;
      auto gradpf_fOld = chOld.gradpf + chOld.gradpf2+2*param.delta*gradpf3Old;

      RF xi3 = -param.tw.Sigma3*(ch.xi/param.tw.Sigma1 + ch.xi2/param.tw.Sigma2);

      RF velocityX = -dxp*ns.vx;
      RF velocityXUpwind = -dxpUpwind*nsUpwind.vx;

      for (size_t i=0; i<ch.pfspace.size(); i++)
      {
        //RF Rspeed = std::max(param.tw.eval_q(ch.pf,ch.pf2,pf3) - 0.1, 0.0);

        RF Rspeed = param.tw.eval_q(ch.pf,ch.pf2,pf3);
        //auto cMod = c -0.2*std::sin(2*3.14152*xpos);
        RF R1 = -Rspeed/param.eps*
                  (param.reaction.eval(c)+param.phys.curvatureAlpha*(ch.xi-xi3))*ch.basis.phi[i];

        //  RF R1 = 0;

        RF J1 = param.mobility.eval(ch.pf,ch.pf2,pf3)*param.eps/param.tw.Sigma1 * (ch.gradxi*ch.basis.gradphi[i][0]);

      // integrate: (pf-pfold)*phi_i/delta_t + gradpf * grad phi_i + ...
        r.accumulate(ch.pfspace,i,factor*(
          (ch.pf-chOld.pf)*ch.basis.phi[i]/param.time.dt
          +J1
          + ( chOld.pf * velocityX - chUpwind.pf * velocityXUpwind) /dx * ch.basis.phi[i]
          + ( (ch.pf * ch.gradvy[0]) + (ch.gradpf[0]*ch.vy) )*ch.basis.phi[i]
          //-pf1*(v*gradphi[i][0])   weak transport
          -R1
        ));

        r.accumulate(ch.pf2space,i,factor*(
          (ch.pf2-chOld.pf2)*ch.basis.phi[i]/param.time.dt
          +param.mobility.eval(ch.pf,ch.pf2,pf3)*param.eps/param.tw.Sigma2 * (ch.gradxi2*ch.basis.gradphi[i][0])
          + ( chOld.pf2 * velocityX - chUpwind.pf2 * velocityXUpwind) /dx * ch.basis.phi[i]
          + ( (ch.pf2 * ch.gradvy[0]) + (ch.gradpf2[0]*ch.vy) )*ch.basis.phi[i]
          //-pf2*(v*gradphi[i][0])
        ));

        r.accumulate(ch.xispace,i,factor*(
          ch.xi*ch.basis.phi[i]
          - 1/param.eps* param.tw.eval_dpf1(ch.pf,ch.pf2,pf3) *ch.basis.phi[i]
          - param.eps * param.tw.Sigma1* (ch.gradpf*ch.basis.gradphi[i][0])
        ));

        r.accumulate(ch.xi2space,i,factor*(
          ch.xi2*ch.basis.phi[i]
          - 1/param.eps* param.tw.eval_dpf2(ch.pf,ch.pf2,pf3) *ch.basis.phi[i]
          - param.eps * param.tw.Sigma2* (ch.gradpf2*ch.basis.gradphi[i][0])
        ));

        //Dune::FieldVector<RF, dim> xg = eg.geometry().global( ip.position() );
        r.accumulate(ch.vyspace,i,factor*(
          ( (pf_f * ch.gradvy[0]) + (gradpf_f*ch.vy) )*ch.basis.phi[i]
          + ( pf_fOld * velocityX - pf_fUpwind * velocityXUpwind) /dx * ch.basis.phi[i]
        ));
      }

    }
  }

};
